/******************************************************************************
 *
 * A class for parsing the 'soundstretch' application command line parameters
 *
 * Author        : Copyright (c) Olli Parviainen 2002
 * Author e-mail : oparviai@iki.fi
 * File created  : 13-Jan-2002
 * Last modified : 29-Jan-2002
 *
 * License :
 *
 *  This file is part of SoundTouch sound processing library.
 *
 *  SoundTouch is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  SoundTouch is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with SoundTouch; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <stdexcept>
#include <string>
#include <stdlib.h>

#include "RunParameters.h"

using namespace std;

// Program usage instructions 

static const char licenseText[] = 
    "    LICENSE:\n"
    "    ========\n"
    "    This program is free software; you can redistribute it and/or modify it\n"
    "    under the terms of the GNU General Public License as published by the\n"
    "    Free Software Foundation; either version 2 of the License, or (at your\n"
    "    option) any later version.\n"
    "\n"
    "    This program is distributed in the hope that it will be useful, but\n"
    "    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY\n"
    "    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License\n"
    "    for more details.\n"
    "\n"
    "    You should have received a copy of the GNU General Public License along\n"
    "    with this program; if not, write to the Free Software Foundation, Inc., 59\n"
    "    Temple Place, Suite 330, Boston, MA  02111-1307 USA\n"
    "\n"
    "This application is distributed with full source codes; however, if you\n"
    "didn't receive them, please visit the author's homepage (see the link above).";

static const char whatText[] = 
    "This application processes WAV audio files by modifying the sound tempo,\n"
    "pitch and playback rate properties independently from each other.\n"
    "\n";

static const char usage[] = 
    "Usage :\n"
    "    soundstretch infile.wav outfile.wav [switches]\n\n"
    "Available switches are:\n"
    "  -tempo=n : change sound tempo by n percents  (n=-95..+5000 %)\n"
    "  -pitch=n : change sound pitch by n semitones (n=-60..+60 semitones)\n"
    "  -rate=n  : change sound rate by n percents   (n=-95..+5000 %)\n"
    "  -quick   : Use quicker tempo change algorithm (gain speed, lose quality)\n"
    "  -naa     : Don't use anti-alias filtering (gain speed, lose quality)\n"
    "  -license : Display the program license text (GPL)\n"
    "\n"
    "Notes :\n"
    "  * The numerical switch parameter values 'n' may be entered either as\n"
    "    integer (e.g. \"123\") or decimal (e.g. \"12.345\") numbers\n"
    "  * The \"-naa\" and/or \"-quick\" switches reduce CPU usage by compromising\n"
    "    some sound quality";




// Converts a char into lower case
static int _toLowerCase(int c)
{
    if (c >= 'A' && c <= 'Z') {
        c += 'a' - 'A';
    }
    return c;
}


// Constructor
RunParameters::RunParameters(const int nParams, const char *paramStr[])
{
    int i;

    if (nParams < 3) {
        // Too few parameters
        if (nParams > 1 && paramStr[1][0] == '-' && 
            _toLowerCase(paramStr[1][1]) == 'l') 
        {
            // '-license' switch
            throwLicense();
        }
        string msg = whatText;
        msg += usage;
        throw runtime_error(msg.c_str());
    }

    inFileName = NULL;
    outFileName = NULL;
    tempoDelta = 0;
    pitchDelta = 0;
    rateDelta = 0;
    quick = 0;
    noAntiAlias = 0;

    // Get input & output file names
    inFileName = (char*)paramStr[1];
    outFileName = (char*)paramStr[2];

    // parse switch parameters
    for (i = 3; i < nParams; i ++) {
        parseSwitchParam(paramStr[i]);
    }

    checkLimits();
}



// Checks parameter limits
void RunParameters::checkLimits()
{
    if (tempoDelta < -95.0f) {
        tempoDelta = -95.0f;
    } else if (tempoDelta > 5000.0f) {
        tempoDelta = 5000.0f;
    }

    if (pitchDelta < -60.0f) {
        pitchDelta = -60.0f;
    } else if (pitchDelta > 60.0f) {
        pitchDelta = 60.0f;
    }

    if (rateDelta < -95.0f) {
        rateDelta = -95.0f;
    } else if (rateDelta > 5000.0f) {
        rateDelta = 5000.0f;
    }
}



// Unknown switch parameter -- throws an exception with an error message
void RunParameters::throwIllegalParamExp(const string &str) const
{
    string msg = "ERROR : Illegal parameter \"";
    msg += str;
    msg += "\".\n\n";
    msg += usage;
    throw runtime_error(msg.c_str());
}



void RunParameters::throwLicense() const
{
    throw runtime_error(licenseText);
}


float RunParameters::parseSwitchValue(const string &str) const
{
    int pos;

    pos = str.find_first_of('=');
    if (pos < 0) {
        // '=' missing
        throwIllegalParamExp(str);
    }

    // Read numerical parameter value after '='
    return atof(str.substr(pos + 1).c_str());
}


// Interprets a single switch parameter string of format "-switch=xx"
// Valid switches are "-tempo=xx", "-pitch=xx" and "-rate=xx". Stores
// switch values into 'params' structure.
void RunParameters::parseSwitchParam(const string &str)
{
    int upS;

    if (str[0] != '-') {
        // leading hyphen missing => not a valid parameter
        throwIllegalParamExp(str);
    }

    // Take the first character of switch name & change to lower case
    upS = _toLowerCase(str[1]);

    // interpret the switch name & operate accordingly
    switch (upS) {
        case 't' :
            // switch '-tempo=xx'
            tempoDelta = parseSwitchValue(str);
            break;

        case 'p' :
            // switch '-pitch=xx'
            pitchDelta = parseSwitchValue(str);
            break;

        case 'r' :
            // switch '-rate=xx'
            rateDelta = parseSwitchValue(str);
            break;

        case 'q' :
            // switch '-quick'
            quick = 1;
            break;

        case 'n' :
            // switch '-naa'
            noAntiAlias = 1;
            break;

        case 'l' :
            // switch '-license'
            throwLicense();
            break;

        default:
            // unknown switch
            throwIllegalParamExp(str);
    }
}
