/******************************************************************************
 *
 * Classes for easy reading & writing of WAV sound files.
 *
 * Many thanks to Manish Bajpai whose WAV writing routines I've used as
 * base of these classes.
 *
 * Author        : Copyright (c) Olli Parviainen 2002
 * Author e-mail : oparviai@iki.fi
 * File created  : 13-Jan-2002
 * Last modified : 21-Feb-2002
 *
 * License :
 *
 *  This file is part of SoundTouch sound processing library.
 *
 *  SoundTouch is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  SoundTouch is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with SoundTouch; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#ifndef WAVFILE_H
#define WAVFILE_H

#include <stdio.h>

#ifndef uint
typedef unsigned int uint;
#endif           

// WAV audio file header structures.

typedef struct {
    char riff_char[4];
    int package_len;
    char wave[4];
} WavRiff;

typedef struct {
    char fmt[4];
    int format_len;
    short int fixed;
    short int channel_number;
    int sample_rate;
    int byte_rate;
    short int byte_per_sample;
    short int bits_per_sample;
} WavFormat;

typedef struct {
    char data_field[4];
    uint data_len;
} WavData;

typedef struct {
    WavRiff riff;
    WavFormat format;
    WavData data;
} WavHeader;



///////////////////////////////////////////////////////////////////////////////
//
// WavInFile : Object for reading WAV audio files
//
class WavInFile
{

private:
    FILE *fptr;
    uint dataRead;
    WavHeader header;

    int readWavHeaders();
    int checkCharTags();
    int readHeaderBlock();
    int readRIFFBlock();

public:
    // Constructor. Opens the .WAV audio file with the given file name. After 
    // constructing the object, the file is ready for reading data or sound 
    // parameters with the functions below.
    //
    // Throws an exception if the file cannot be opened.
    WavInFile(const char *filename);

    // Destructor. Closes the audio file.
    ~WavInFile();

    // Closes the file.
    void close();

    // returns sample rate of the audio file.
    uint getSampleRate() const;

    // returns number of bits per sample of the audio file.
    uint getNumBits() const;

    // returns the total number of sound data bytes.
    uint getDataSizeInBytes() const;

    // returns the total number of sound samples in the file.
    uint getNumSamples() const;

    // returns number of bytes used for a sample, e.g. 16 bit stereo = 4 bytes
    uint getBytesPerSample() const;

    // returns number of sound channels of the audio file.
    uint getNumChannels() const;

    // returns the total file audio length in milliseconds.
    uint getLengthMS() const;

    // Reads sound data from the file to the given bufffer.
    // Returns number of bytes read.
    int read(void *buffer, const int bufferSizeInBytes);

    // Returns nonzero if end of the file has been reached.
    int eof() const;
};


///////////////////////////////////////////////////////////////////////////////
//
// WavOutFile : Object for writing WAV audio files
//
class WavOutFile
{
private:
    FILE *fptr;
    WavHeader header;
    int bytesWritten;

    void fillInHeader(const uint sampleRate, const uint bits, const uint channels);
    void finishHeader();
    void writeHeader();

public:
    // Constructor. Creates a new WAV audio file ready for writing. After 
    // constructing, the file is ready for writing sound data.
    //
    // Throws an exception if the file couldn't be created.
    WavOutFile(const char *fileName, const int sampleRate, const int bits, const int channels);

    // Destructor. Finishes and closes the WAV audio file.
    ~WavOutFile();

    // Writes audio data into the file.
    void write(const void *buffer, const int numBytes);

    // Closes the file.
    void close();
};

#endif

